@@ -20,6 +20,7 @@
 #include "leafnode.h"
 #include "activutil.h"
 #include "ln_log.h"
+#include "mastring.h"
 
 #include <ctype.h>
 #include "system.h"
@@ -66,7 +67,7 @@ compactive(const void *a, const void *b)
     return strcasecmp(la->name, lb->name);
 }
 
-struct nglist *newgroup = NULL;
+static struct nglist *newgroup = NULL;
 
 /*
  * insert a group into a list of groups
@@ -119,11 +120,28 @@ insertgroup(const char *name, long unsigned first,
     lold = l;
 }
 
+void
+newgroupdesc(const char *groupname, const char *description)
+{
+    struct nglist *l = newgroup;
+
+    while(l) {
+	if (0 == strcasecmp(groupname, l->entry->name)) {
+	    if (l->entry->desc)
+		free(l->entry->desc);
+	    l->entry->desc = critstrdup(description, "newgroupdesc");
+	    break;
+	}
+	l = l->next;
+    }
+}
+
+
 /*
  * change description of newsgroup
  */
 void
-changegroupdesc(const char *groupname, char *description)
+changegroupdesc(const char *groupname, const char *description)
 {
     struct newsgroup *ng;
 
@@ -252,18 +270,16 @@ writeactive(void)
 {
     FILE *a;
     struct newsgroup *g;
-    char *c;
+    mastr *c = mastr_new(PATH_MAX), *d;
     int err;
     size_t count = 0;
-    char s[SIZE_s+1];
 
-    c = critmalloc(PATH_MAX, "Allocating buffer");
-
-    xsnprintf(s, SIZE_s, "%s/leaf.node/groupinfo.new", spooldir);
-    a = fopen(s, "w");
+    mastr_vcat(c, spooldir, "/leaf.node/groupinfo.new", NULL);
+    (void)unlink(mastr_str(c));
+    a = fopen(mastr_str(c), "w");
     if (!a) {
-	ln_log(LNLOG_SERR, LNLOG_CTOP, "cannot open new groupinfo file \"%s\": %m", s);
-	free(c);
+	ln_log(LNLOG_SERR, LNLOG_CTOP, "cannot open new groupinfo file \"%s\": %m", mastr_str(c));
+	mastr_delete(c);
 	return -1;
     }
 
@@ -303,17 +319,19 @@ writeactive(void)
 
     if (err == EOF) {
 	ln_log(LNLOG_SERR, LNLOG_CTOP,
-		"failed writing new groupinfo file \"%s\": %m", s);
-	unlink(s);
-	free(c);
+		"failed writing new groupinfo file \"%s\": %m", mastr_str(c));
+	unlink(mastr_str(c));
+	mastr_delete(c);
 	return -1;
     }
 
-    xsnprintf(c, PATH_MAX, "%s/leaf.node/groupinfo", spooldir);
-    if (rename(s, c)) {
-	ln_log(LNLOG_SERR, LNLOG_CTOP, "failed to rename new groupinfo \"%s\" file into proper place \"%s\": %m", s, c);
-	unlink(s);
-	free(c);
+    d = mastr_new(PATH_MAX);
+    mastr_vcat(d, spooldir, "/leaf.node/groupinfo", NULL);
+    if (rename(mastr_str(c), mastr_str(d))) {
+	ln_log(LNLOG_SERR, LNLOG_CTOP, "failed to rename new groupinfo \"%s\" file into proper place \"%s\": %m", mastr_str(c), mastr_str(d));
+	unlink(mastr_str(c));
+	mastr_delete(d);
+	mastr_delete(c);
 	return -1;
     } else {
 	if (verbose) printf("wrote active file with %lu lines\n",
@@ -321,7 +339,8 @@ writeactive(void)
 	syslog(LOG_INFO, "wrote active file with %lu lines",
 	       (unsigned long)count);
     }
-    free(c);
+    mastr_delete(d);
+    mastr_delete(c);
     return 0;
 }
 
@@ -388,8 +407,12 @@ readactive(void)
 	    }
 	    fclose(f);
     } else {
+	    int e = errno;
 	    syslog(LOG_ERR, "unable to open %s: %m", s);
-	    return;
+	    if (e == ENOENT)
+		return;
+	    fprintf(stderr, "unable to open %s: %s, aborting", s, strerror(e));
+	    exit(1);
     }
 
     n = ((off_t) n > st.st_size) ? st.st_size : (off_t) n;
@@ -460,7 +483,7 @@ readactive(void)
 	if (strcmp(r, "-x-") == 0)
 	    g->desc = NULL;
 	else
-	    g->desc = strdup(r);
+	    g->desc = critstrdup(r, "readactive");
 
 	p = q;			/* next record */
 	g++;
@@ -500,7 +523,10 @@ fakeactive(void)
     long unsigned first, last;
     char *p;
     char s[SIZE_s+1];
+    struct stat st;
+    time_t age;
 
+    killactiveread(); /* force reading active file regardless */
     xsnprintf(s, SIZE_s, "%s/interesting.groups", spooldir);
     d = opendir(s);
     if (!d) {
@@ -541,13 +567,39 @@ fakeactive(void)
 	    if (debugmode)
 		syslog(LOG_DEBUG, "parsed directory %s: first %lu, last %lu",
 		       de->d_name, first, last);
-	    /* as this is fake and we don't want old groups reported as
-	     * new to clients doing NEWGROUPS, age 0 is OK here
-	     */
-	    insertgroup(de->d_name, first, last, 0);
+	    if (0 == stat(".", &st))
+		age = st.st_ctime;
+	    else
+		age = 0;
+	    insertgroup(de->d_name, first, last, age);
 	}
     }
     mergegroups();
 
     closedir(d);
 }
+
+char *
+activeread(void)
+{
+    const char *a = "/active.read";
+    size_t l;
+    char *t = critmalloc((l = strlen(spooldir)) + strlen(a) + 1, "activeread");
+    strcpy(t, spooldir); /* RATS: ignore */
+    strcpy(t + l, a); /* RATS: ignore */
+    return t;
+}
+
+int
+killactiveread(void)
+{
+    int rc = 0;
+    char *t = activeread();
+    if (unlink(t) && errno != ENOENT) {
+	ln_log(LNLOG_SERR, LNLOG_CTOP,
+		"cannot delete %s: %m", t);
+	rc = -1;
+    }
+    free(t);
+    return rc;
+}
